@@ -2,7 +2,7 @@
 # vc-dwim - a version-control-agnostic ChangeLog diff and commit tool
 # @configure_input@
 #
-# Copyright 2006-2017 Free Software Foundation, Inc.
+# Copyright 2006-2019 Free Software Foundation, Inc.
 # Written by Jim Meyering <meyering@redhat.com>
 
 # This program is free software: you can redistribute it and/or modify
@@ -16,7 +16,7 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 # Given a command like vc-dwim ChangeLog lib/ChangeLog..., check that each
 # ChangeLog has been modified, determine the list of affected files from
@@ -39,6 +39,7 @@ BEGIN
 {
   my $perllibdir = $ENV{'perllibdir'} || '@datadir@/@PACKAGE@';
   unshift @INC, (split '@PATH_SEPARATOR@', $perllibdir);
+  unshift @INC, dirname $0; # DELETE_ME upon installation
 
   # Override SHELL.  This is required on DJGPP so that system() uses
   # bash, not COMMAND.COM which doesn't quote arguments properly.
@@ -60,6 +61,7 @@ our $VERSION = '@VERSION@';
 
 my $verbose = 0;
 my $debug = 0;
+my $dry_run = 0;
 
 sub usage ($)
 {
@@ -114,6 +116,10 @@ sub get_diffs ($$$)
   my ($vc, $f, $pristine) = @_;
 
   my @cmd = ($pristine ? $vc->diff_pristine() : $vc->diff_cmd(), @$f);
+  if ($dry_run) {
+    print "$ME: would run: @cmd\n";
+    return [];
+  }
   $verbose
     and verbose_cmd \@cmd;
   open PIPE, '-|', @cmd
@@ -150,6 +156,7 @@ sub get_new_changelog_lines ($$)
   my $diff_lines = get_diffs ($vc, [$f], 1);
   if (@$diff_lines == 0)
     {
+      $dry_run and return [];
       my $vc_name = $vc->name();
       die qq|$ME: "$vc_name diff $f" produced no output\n|;
     }
@@ -247,7 +254,7 @@ sub run_command
      IGNORE_FAILURE => 0,
      DIE_UPON_FAILURE => 1,
      INHIBIT_STDERR => 0,
-     INHIBIT_STDOUT => 1,
+     INHIBIT_STDOUT => ! $dry_run, # keep stdout if just printing
     );
 
   my %options = %all_options;
@@ -291,7 +298,13 @@ sub run_command
     }
 
   my $fail = 1;
-  my $rc = 0xffff & system @cmd;
+  my $rc;
+  if ($dry_run) {
+    print "$ME: would run: @cmd\n";
+    $rc = 0;
+  } else {
+    $rc = 0xffff & system @cmd;
+  }
 
   # Restore stdout.
   open STDOUT, '>&', *SAVE_OUT
@@ -549,6 +562,8 @@ sub do_commit ($$$$)
 sub do_at($$)
 {
   my ($dest_dir, $code) = @_;
+  eval 'use Cwd';
+  die $@ if $@;
   my $initial_wd = Cwd::getcwd()
     or die "$ME: getcwd failed: $!\n";
   chdir $dest_dir
@@ -690,11 +705,24 @@ sub check_attribution($$)
     or die "$ME: --author/ChangeLog mismatch:\n  $$author\n  $name_and_email\n";
 }
 
-# Return the name of the/an admin directory residing in
+# Return the name of the/an admin directory associated with
 # the current directory.  If there is none, return undef.
 sub admin_dir()
 {
+  # This is the usual case: a .git directory.
   -d '.git/objects' and return '.git';
+  # With a git worktree, .git is a file containing a line of this form:
+  # gitdir: /abs/dir
+  my $git_file = '.git';
+  if (-f $git_file)
+    {
+      # Read .git, and extract the name after "gitdir: "
+      my $fh = new IO::File $git_file, 'r'
+        or die "$ME: can't open `$git_file' for reading: $!\n";
+      my $line = <$fh>;
+      defined $line && $line =~ /^gitdir: (.+)$/
+        and return $1;
+    }
   -d '.hg' and return '.hg';
   -d 'CVS' and return 'CVS';
   -d '.svn' and return '.svn';
@@ -718,6 +746,8 @@ sub main
      'author=s' => \$author,   # makes sense only with --commit
      'print-vc-list' =>
        sub { print join (' ', VC::supported_vc_names()), "\n"; exit },
+     n => \$dry_run,
+     'dry-run' => \$dry_run,
      debug => \$debug,
      verbose => \$verbose,
      help => sub { usage 0 },
@@ -738,34 +768,51 @@ sub main
          INHIBIT_STDOUT => 0,
         };
 
-      # set -e; cd $adm && mkdir -p c && cd c
-      chdir $adm or die "$ME: failed to chdir to $adm: $!\n";
-      ! (mkdir ('c') || $! == EEXIST)
-        and die "$ME: failed to create $adm/c: $!\n";
-      chdir 'c' or die "$ME: failed to chdir to $adm/c: $!\n";
-
-      # touch ChangeLog || die
       my $cl = 'ChangeLog';
-      open FH, '>>', $cl
-        or die "$ME: failed to open '$cl' for writing: $!\n";
-      close FH
-        or die "$ME: failed to write '$cl': $!\n";
+      do_at ($adm, sub
+      {
+        if ($dry_run) {
+          print "$ME: would mkdir 'c' in $adm\n";
+        } else {
+          ! (mkdir ('c') || $! == EEXIST)
+            and die "$ME: failed to create $adm/c: $!\n";
+          chdir 'c' or die "$ME: failed to chdir to $adm/c: $!\n";
+        }
 
-      # Initialize the git repo, add ChangeLog and commit it.
-      # Any failure is fatal.
-      run_command ($options, qw(git init -q));
-      run_command ($options, qw(git add), $cl);
-      run_command ($options, qw(git commit --allow-empty -q -m. -a));
+        # touch ChangeLog || die
+        if ($dry_run) {
+          print "$ME: would touch $cl in $adm\n";
+        } else {
+          open FH, '>>', $cl
+            or die "$ME: failed to open '$cl' for writing: $!\n";
+          close FH
+            or die "$ME: failed to write '$cl': $!\n";
+        }
 
-      # If the top-level ChangeLog file exists, rename it to ChangeLog~,
-      # deliberately ignoring any rename failure.
-      my $cl_top = "../../$cl";
-      rename $cl_top, "$cl_top~";
+        # Initialize the git repo, add ChangeLog and commit it.
+        # Any failure is fatal.
+        run_command ($options, qw(git init -q));
+        run_command ($options, qw(git add), $cl);
+        run_command ($options, qw(git commit --allow-empty -q -m. -a));
+      });
+
+      # If a ChangeLog file exists in the current directory, rename it
+      # deliberately ignoring any rename failure. (But only report the
+      # rename for dry runs if it does exist.)
+      if ($dry_run) {
+        -e $cl and print "$ME: would rename($cl, $cl~)\n";
+      } else {
+        rename $cl, "$cl~";
+      }
 
       # Create the top-level ChangeLog symlink into $adm/c:
       my $cl_sub = "$adm/c/$cl";
-      symlink $cl_sub, $cl_top
-        or die "$ME: failed to create symlink, $cl, to $cl_sub: $!\n";
+      if ($dry_run) {
+        print "$ME: would symlink($cl_sub, $cl)\n";
+      } else {
+        symlink $cl_sub, $cl
+          or die "$ME: failed to create symlink, $cl, to $cl_sub: $!\n";
+      }
 
       exit 0;
     }
@@ -783,7 +830,12 @@ sub main
       $verbose
         and verbose_cmd \@cmd;
 
-      exec @cmd;
+      if ($dry_run) {
+        print "$ME: would run: @cmd\n";
+        exit 0;
+      } else {
+        exec @cmd;
+      }
       exit 1;
     }
 
@@ -881,6 +933,9 @@ sub main
           $added_log_lines{$log} = get_new_changelog_lines $vc, $log;
         }
     }
+  $dry_run
+    and exit 0;
+
   foreach my $log (@changelog_file_name)
     {
       my $line_list = $added_log_lines{$log};
@@ -1218,14 +1273,14 @@ Print the list of recognized version control names, then exit.
 
 =item B<--initialize>
 
-This option is useful in a project that does not version-control
-a ChangeLog file.  Use this option in the top-level project
-directory to create your personal ChangeLog file -- that file
-will be a symlink to a git-version-controlled ChangeLog file
-in a just-created single-file repository residing in the VC-admin
-directory (.git, .hg, etc.).  If there is an existing ChangeLog file
-in the working directory, running B<vc-dwim --initialize> first
-renames it to ChangeLog~.
+This option, or the equivalent operations, is needed in a project that
+does not version-control a ChangeLog file. Use this option in the
+top-level project directory to create your personal ChangeLog file --
+that file will be a symlink to a git-version-controlled ChangeLog file
+in a just-created single-file repository residing in the VC admin
+directory (.git, .hg, etc.). If there is an existing C<ChangeLog> file
+in the top-level directory, running B<vc-dwim --initialize> first
+renames it to C<ChangeLog~>.
 
 =item B<--help>
 
@@ -1259,7 +1314,11 @@ Eventually, it will detect temporaries created by other editors.
 
 Jim Meyering <jim@meyering.net>
 
-Report bugs to <@PACKAGE_BUGREPORT@>.
+Report bugs and all other discussion to <@PACKAGE_BUGREPORT@>.
+
+GNU <@PACKAGE_NAME> home page: &lt;https://www.gnu.org/software/<@PACKAGE_NAME>/&gt;
+
+General help using GNU software: <https://www.gnu.org/gethelp/>
 
 =cut
 
